from __future__ import absolute_import

import os
import unittest
from operator import add

from uniRW.File import DataFile
from uniRW.HReader import HReader
from uniRW.Line import Line
from uniRW.State import State
from uniRW.Value import Value, StateValue
from uniRW.util import pureR, pureL


class TestReader(unittest.TestCase):

    def setUp(self):

        def average(num_list):
            return round(sum(num_list) / len(num_list), 2)

        name = Value('Name')
        course = Value('Course')
        number = StateValue('Number')
        grade1 = Value("Grade", map_f=pureR(float), reduce_f=max, post_reduce_f=max)
        grade2 = Value(
            name='Grade',
            map_f=pureR(lambda x: [float(x)]),
            reduce_f=add,
            post_map_f=pureR(average),
            post_reduce_f=max,
        )

        example1 = os.path.join(os.path.dirname(__file__), 'data/example1.csv')
        example2A = os.path.join(os.path.dirname(__file__), 'data/example2A.csv')
        example2B = os.path.join(os.path.dirname(__file__), 'data/example2B.csv')

        self.example1 = DataFile(example1, Line(','), header_lineno=1)
        self.example2A = DataFile(example2A, Line(','), header_lineno=0)
        self.example2B = DataFile(example2B, Line(','), header_lineno=0)

        self.hierarchy11 = {name: [grade1]}
        self.hierarchy12 = {name: [grade2]}
        self.hierarchy13 = {name: [grade1, number]}
        self.hierarchy21 = {name: {course: [grade1]}}
        self.hierarchy22 = {name: {course: [grade2]}}

    def test_read_lines(self):
        value_dict11 = {'Alice': {'Grade': 4.0}}
        value_dict12 = {'Alice': {'Grade': 4.0}, 'Bob': {'Grade': 3.0}}
        value_dict21 = {'Alice': {'Math': {'Grade': 4.0}, 'CS': {'Grade': 3.9}}}
        value_dict22 = {'Alice': {'Math': {'Grade': 4.0}, 'CS': {'Grade': 3.8}},
                        'Bob': {'Math': {'Grade': 3.0}, 'CS': {'Grade': 3.9}}}
        value_lines11 = [{'Name': 'Alice', 'Grade': 4.0}]
        value_lines12 = [{'Name': 'Alice', 'Grade': 4.0}, {'Name': 'Bob', 'Grade': 3.0}]
        value_lines21 = [{'Name': 'Alice', 'Course': 'Math', 'Rank': 1, 'Grade': 4.0},
                         {'Name': 'Alice', 'Course': 'CS', 'Rank': 2, 'Grade': 3.9}]
        value_lines22 = [{'Name': 'Alice', 'Course': 'Math', 'Rank': 1, 'Grade': 4.0},
                         {'Name': 'Alice', 'Course': 'CS', 'Rank': 2, 'Grade': 3.8},
                         {'Name': 'Bob', 'Course': 'Math', 'Rank': 2, 'Grade': 3.0},
                         {'Name': 'Bob', 'Course': 'CS', 'Rank': 1, 'Grade': 3.9}]

        grade_dict11, _ = HReader(self.hierarchy11).read_lines(value_lines11)
        grade_dict12, _ = HReader(self.hierarchy11).read_lines(value_lines12)
        grade_dict21, _ = HReader(self.hierarchy21).read_lines(value_lines21)
        grade_dict22, _ = HReader(self.hierarchy21).read_lines(value_lines22)

        self.assertEqual(grade_dict11, value_dict11)
        self.assertEqual(grade_dict12, value_dict12)
        self.assertEqual(grade_dict21, value_dict21)
        self.assertEqual(grade_dict22, value_dict22)

    def test_read(self):
        def skip_first_line(line):
            return not line[0][0] == '#'

        def update(state):
            state['Number'] += 1

        state = State({'Number': 0}, pureL(update))

        grade_dict11, _ = HReader(self.hierarchy11, filter_f=pureR(skip_first_line)).read(self.example1)
        grade_dict12, _ = HReader(self.hierarchy12, filter_f=pureR(skip_first_line)).read(self.example1)
        grade_dict13, final_state = HReader(self.hierarchy13, state, pureR(skip_first_line)).read(self.example1)

        grade_dict21A, _ = HReader(self.hierarchy21).read(self.example2A)
        grade_dict21B, _ = HReader(self.hierarchy21).read(self.example2B)
        grade_dict21C, _ = HReader(self.hierarchy21, state).read(self.example2A, carry_state=True)
        grade_dict22A, _ = HReader(self.hierarchy22).read(self.example2A)
        grade_dict22B, _ = HReader(self.hierarchy22).read(self.example2B)
        grade_dict22AP, _ = HReader(self.hierarchy22).read(self.example2A, apply_post_map=True)
        grade_dict22BP, _ = HReader(self.hierarchy22).read(self.example2B, apply_post_map=True)

        self.assertEqual(grade_dict11, {'Alice': {'Grade': 3.5}, 'Bob': {'Grade': 4.0}})
        self.assertEqual(grade_dict12, {'Alice': {'Grade': [3.5, 3.4]}, 'Bob': {'Grade': [2.6, 4.0]}})
        self.assertEqual(grade_dict13, {'Alice': {'Grade': 3.5, 'Number': 5}, 'Bob': {'Grade': 4.0, 'Number': 4}})
        self.assertEqual(final_state['Number'], 5)

        self.assertEqual(grade_dict21A, {'Alice': {'Math': {'Grade': 100.0}}, 'Bobby': {'Math': {'Grade': 90.0}}})
        self.assertEqual(grade_dict21B, {'Alice': {'CS': {'Grade': 95.0}}, 'Bobby': {'CS': {'Grade': 100.0}}})
        self.assertEqual(state['Number'], 7)
        self.assertEqual(grade_dict22A, {'Alice': {'Math': {'Grade': [100.0, 90.0, 95.0]}},
                                         'Bobby': {'Math': {'Grade': [60.0, 90.0, 70.0, 80.0]}}})
        self.assertEqual(grade_dict22B, {'Alice': {'CS': {'Grade': [80.0, 90.0, 75.0, 95.0]}},
                                         'Bobby': {'CS': {'Grade': [100.0, 90.0, 95.0]}}})
        self.assertEqual(grade_dict22AP, {'Alice': {'Math': {'Grade': 95.0}},
                                          'Bobby': {'Math': {'Grade': 75.0}}})
        self.assertEqual(grade_dict22BP, {'Alice': {'CS': {'Grade': 85.0}},
                                          'Bobby': {'CS': {'Grade': 95.0}}})

    def test_read_all(self):
        grade_dict11 = HReader(self.hierarchy11).read_all([self.example2A, self.example2B])
        grade_dict12 = HReader(self.hierarchy12).read_all([self.example2A, self.example2B])
        grade_dict21 = HReader(self.hierarchy21).read_all([self.example2A, self.example2B])
        grade_dict22 = HReader(self.hierarchy22).read_all([self.example2A, self.example2B])

        self.assertEqual(grade_dict11, {'Alice': {'Grade': 100.0}, 'Bobby': {'Grade': 100.0}})
        self.assertEqual(grade_dict12, {'Alice': {'Grade': 95.0}, 'Bobby': {'Grade': 95.0}})
        self.assertEqual(grade_dict21, {'Alice': {'Math': {'Grade': 100.0}, 'CS': {'Grade': 95.0}},
                                        'Bobby': {'Math': {'Grade': 90.0}, 'CS': {'Grade': 100.0}}})
        self.assertEqual(grade_dict22, {'Alice': {'Math': {'Grade': 95.0}, 'CS': {'Grade': 85.0}},
                                        'Bobby': {'Math': {'Grade': 75.0}, 'CS': {'Grade': 95.0}}})


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function

from uniRW.util import null


class State:

    def __init__(self, init_state, update_func=null):
        """Initialize a state for stateful file reading.

        :param init_state (dict): a dictionary (name => value) storing the initial state.
        :param update_func (State * Line -> ()): a function that update state after reading each line.
        """
        self.__state = init_state
        self.update_func = update_func
        self.__read_only = True

    def __getitem__(self, name):
        return self.get(name)

    def __setitem__(self, name, value):
        return self.set(name, value)

    def __contains__(self, name):
        return name in self.__state

    def get(self, name):
        if name not in self.__state:
            raise KeyError(str(name) + " is not in state")
        else:
            return self.__state[name]

    def set(self, name, val):
        if not self.__read_only:
            if name not in self.__state:
                raise KeyError(str(name) + " is not in state")
            else:
                self.__state[name] = val
        else:
            raise KeyError("Error: cannot change state")

    def lock(self):
        self.__read_only = True

    def release(self):
        self.__read_only = False

    def update(self, val):
        if not self.__read_only:
            self.update_func(self, val)
        else:
            raise KeyError("Error: cannot change state")

    def print(self):
        print(self.__state)



from __future__ import absolute_import
import unittest

import os
from shutil import rmtree

from uniRW.File import OutputLine
from uniRW.File import OutputFile
from uniRW.Value import Value
from uniRW import HWriter


class TestWriter(unittest.TestCase):

    def setUp(self):
        output = os.path.join(os.path.dirname(__file__), 'output')
        os.mkdir(output)

    def tearDown(self):
        output = os.path.join(os.path.dirname(__file__), 'output')
        rmtree(output)

    def test_write(self):
        name = Value('Name')
        course = Value('Course')
        rank = Value('Rank')
        grade1 = Value("Grade")
        grade2 = Value("Grade", to_string=lambda x: str(max(x)))
        grade3 = Value("Grade", to_string=lambda x: str(x[0]) + ',' + str(x[1]))

        hierarchy1 = {name: {course: [rank, grade1]}}
        hierarchy2 = {name: {course: [rank, grade2]}}
        hierarchy3 = {name: {course: [grade3]}}

        output_file1 = os.path.join(os.path.dirname(__file__), 'output/output1.csv')
        output_file2 = os.path.join(os.path.dirname(__file__), 'output/output2.csv')
        output_file3 = os.path.join(os.path.dirname(__file__), 'output/output3.csv')

        value_line1 = [name, course, rank, grade1]
        value_line2 = [name, course, rank, grade2]
        value_line3 = [name, course, grade3]

        output1 = OutputFile(output_file1, OutputLine(','), epilogue=['# Test output', '# uniRW HWriter'])
        output2 = OutputFile(output_file2, OutputLine(','), foreword=['# Test output', '# uniRW HWriter'])
        output3 = OutputFile(
            output_file3,
            OutputLine(','),
            header=['Name', 'Course', 'Rank', 'Grade']
        )

        value_dict1 = {'Alice': {'Math': {'Rank': 1, 'Grade': 100.0}, 'CS': {'Rank': 2, 'Grade': 95.0}},
                       'Bobby': {'Math': {'Rank': 2, 'Grade': 90.0}, 'CS': {'Rank': 1, 'Grade': 98.0}}}
        value_dict2 = {'Alice': {'Math': {'Rank': 1, 'Grade': [100.0, 90.0, 95.0]},
                                 'CS': {'Rank': 2, 'Grade': [80.0, 90.0, 75.0]}},
                       'Bobby': {'Math': {'Rank': 2, 'Grade': [60.0, 70.0, 80.0]},
                                 'CS': {'Rank': 1, 'Grade': [90.0, 95.0]}}}
        value_dict3 = {'Alice': {'Math': {'Grade': (1, 100.0)}, 'CS': {'Grade': (2, 95.0)}},
                       'Bobby': {'Math': {'Grade': (2, 90.0)}, 'CS': {'Grade': (1, 95.0)}}}

        HWriter(hierarchy1, value_line1).write(output1, value_dict1, sort_by='Grade')
        HWriter(hierarchy2, value_line2).write(output2, value_dict2, sort_by='Grade')
        HWriter(hierarchy3, value_line3).write(output3, value_dict3, sort_by='Grade')

        ref_file1 = os.path.join(os.path.dirname(__file__), 'data/output1.csv')
        ref_file2 = os.path.join(os.path.dirname(__file__), 'data/output2.csv')
        ref_file3 = os.path.join(os.path.dirname(__file__), 'data/output3.csv')

        ref1 = open(ref_file1, 'r')
        ref2 = open(ref_file2, 'r')
        ref3 = open(ref_file3, 'r')

        out1 = open(output_file1, 'r')
        out2 = open(output_file2, 'r')
        out3 = open(output_file3, 'r')

        self.assertEqual(ref1.read(), out1.read())
        self.assertEqual(ref2.read(), out2.read())
        self.assertEqual(ref3.read(), out3.read())

        ref1.close()
        ref2.close()
        ref3.close()
        out1.close()
        out2.close()
        out3.close()


if __name__ == '__main__':
    unittest.main()



from re import split as split


class Line:

    def __init__(self, delimiter, name_col_dict={}):
        """Initialize a line object for reading.

        :param delimiter (str): a regular expression that splits a raw data line.
        :param name_col_dict (dict): a dictionary (name => column) that maps value name to column number.
        """
        self.name_col_dict = name_col_dict.copy()
        self.delimiter = delimiter
        self.__line = []

    def __getitem__(self, item):
        if type(item) is int:
            return self.get_by_index(item)
        else:
            return self.get_by_name(item)

    def __contains__(self, name):
        return name in self.name_col_dict

    def get_by_name(self, name):
        if name in self.name_col_dict:
            return self.__line[self.name_col_dict[name]]
        else:
            raise KeyError(str(name) + " is not in a data line")

    def get_by_index(self, index):
        return self.__line[index]

    def store(self, line):
        """convert store a raw data line into a list and store it in the line object

        :param line (str): a raw data line
        """
        self.__line = split(self.delimiter, line)

    def set_header(self):
        """Make the name => column dictionary from current stored values by indexing.
        """
        for i, val in enumerate(self.__line):
            if val in self.name_col_dict:
                continue
            else:
                self.name_col_dict[val] = i

    def clear(self):
        self.name_col_dict = {}
        self.__line = []


class OutputLine:

    def __init__(self, delimiter, start_char='', end_char='\n'):
        """Initialize a outputLine object for writing.

        :param delimiter (str): the delimiter of a line.
        :param end_char: the ending character of a line.
        """
        self.delimiter = delimiter
        self.start_char = start_char
        self.end_char = end_char

    def get_line(self, values):
        """Join a list of strings by the delimiter

        :param values ([str]): a list of strings
        :return (str): the result string
        """
        return self.start_char + self.delimiter.join(values) + self.end_char



from __future__ import print_function
from __future__ import absolute_import
from uniRW.util import idenR


class GeneralKey:

    def __init__(self, name, map_f, to_string):
        self.name = name
        self.map_f = map_f
        self.to_string = to_string


class Key(GeneralKey):

    def __init__(self, name, map_f=idenR, to_string=str):
        GeneralKey.__init__(self, name, map_f, to_string)


class StateKey(GeneralKey):

    def __init__(self, name, map_f=idenR, to_string=str):
        GeneralKey.__init__(self, name, map_f, to_string)



from __future__ import absolute_import

from copy import copy

from uniRW.File import DataFile
from uniRW.Value import Value, StateValue
from uniRW.deprecated.Key import Key, StateKey, GeneralKey


class Reader:

    def __init__(self, Key, Values, State=None, filter_f=lambda dummy,x:True):
        if isinstance(Key, GeneralKey):
            self.Key = Key
        else:
            raise ValueError("Key is not a Key or StateKey object")
        self.Values = Values
        self.State = State
        self.filter_f = filter_f

    def read(self, data_file, mode='r', apply_post_map=False):

        lineno = 0
        key_val_dict = {}
        current_state = None
        if self.State != None:
            current_state = copy(self.State)

        if not isinstance(data_file, DataFile):
            raise ValueError("Data file is not a DataFile object.")

        with open(data_file.file_name, mode) as file:

            for line in file:

                data_file.line.store(line[:-1])

                if data_file.header_lineno == lineno:
                    data_file.line.set_header()
                    lineno += 1
                    continue

                if self.State != None:
                    current_state.release()
                    current_state.update(data_file.line)
                    current_state.lock()

                if not self.filter_f(current_state, data_file.line):
                    lineno += 1
                    continue

                if isinstance(self.Key, Key):
                    key = self.Key.map_f(current_state, data_file.line.get_by_name(self.Key.name))
                elif isinstance(self.Key, StateKey):
                    key = self.Key.map_f(current_state, current_state.get(self.Key.name))


                for value in self.Values:

                    val_name = value.name

                    if isinstance(value, StateValue):
                      if current_state.check(val_name):
                          val = value.map_f(current_state, current_state.get(val_name))
                      else:
                        raise KeyError(str(val_name) + " is not in state")
                    elif isinstance(value, Value):
                        val = value.map_f(current_state, data_file.line.get_by_name(val_name))
                    else:
                        raise ValueError("Value is not a Value or StateValue object")

                    if key in key_val_dict:
                        if val_name in key_val_dict[key]:
                            current_val = key_val_dict[key][val_name]
                            reduced_val = value.reduce_f(current_val, val)
                            key_val_dict[key][val_name] = reduced_val

                        else:
                            key_val_dict[key][val_name] = val
                    else:
                        key_val_dict[key] = {}
                        key_val_dict[key][val_name] = val

                lineno += 1

        def post_map(state, v_dict):
            new_v_dict = v_dict
            for value in self.Values:
                val_name = value.name
                current_val = v_dict[val_name]
                new_v_dict[val_name] = value.post_map_f(state, current_val)
            return new_v_dict

        if apply_post_map:
            final_key_val_dict = {k: post_map(current_state,v_dict) for k, v_dict in key_val_dict.items()}
            return final_key_val_dict, current_state
        else:
            return key_val_dict, current_state


    def readAll(self, data_files, mode='r'):

        final_key_val_dict = {}

        for data_file in data_files:
            key_val_dict, final_state = self.read(data_file= data_file, mode=mode)

            for key, val_dict in key_val_dict.items():

                if key in final_key_val_dict:

                    for value in self.Values:

                        val_name = value.name
                        raw_val = val_dict[val_name]
                        val = value.post_map_f(final_state, raw_val)

                        if val_name in final_key_val_dict[key]:
                            current_val = final_key_val_dict[key][val_name]
                            reduced_val = value.post_reduce_f(current_val, val)
                            final_key_val_dict[key][val_name] = reduced_val
                        else:
                            final_key_val_dict[key][val_name] = val

                else:
                    final_key_val_dict[key] = {}

                    for value in self.Values:

                        val_name = value.name
                        raw_val = val_dict[val_name]
                        val = value.post_map_f(final_state, raw_val)
                        final_key_val_dict[key][val_name] = val

        return final_key_val_dict



from __future__ import absolute_import

from copy import deepcopy, copy
from os.path import isfile

from uniRW.Line import Line, OutputLine


class File:

    def __init__(self, file_name, line):
        """Initialize a file object.

        :param file_name (str): name.
        :param line (Line or OutputLine): a Line or OutputLine object describing a line in this file.
        """
        self.file_name = file_name
        self.line = deepcopy(line)  # prevent bugs in sharing the same Line object


class DataFile(File):

    def __init__(self, file_name, line, header_lineno=-1):
        """Initialize a file object for reading.

        :param header_lineno (int): the line number (starting from 0) of the header of the file
        """
        if not isinstance(line, Line):
            raise ValueError("line is not a Line object.")
        File.__init__(self, file_name, line)
        if not isfile(file_name):
            raise ValueError(str(file_name) + " is not a file or not found")
        self.header_lineno = header_lineno
        self.__init_line = deepcopy(line)  # private copy of line

    def reset(self):
        self.line = deepcopy(self.__init_line)


class OutputFile(File):

    def __init__(self, file_name, line, header=[], foreword=[], epilogue=[]):
        """Initialize a file object for writing.

        :param header ([str]): the header of file.
        :param foreword ([str]): the foreword of a file with each string as a line.
        :param epilogue ([str]): the epilogue of a file with each string as a line.
        """
        if not isinstance(line, OutputLine):
            raise ValueError("line is not a OutputLine object.")
        File.__init__(self, file_name, line)
        self.header = copy(header)
        self.foreword = copy(foreword)
        self.epilogue = copy(epilogue)



from __future__ import absolute_import

from copy import deepcopy

from uniRW.State import State
from uniRW.Value import GeneralValue, Value, StateValue


class Hierarchy:
    """A hierarchy specification of values defines the format of data being read or written.

    Type 1 (list): represents values in the same layer (usually at the bottom)
    Type 2 (dict): represents the value in the current layer and sub-hierarchy of next layer
    Type 3 (tuple): similar to Type 2 (not recommended because of lack of testing)

    A valid hierarchy specification would be a composition of the above three types following these Specs:

    Spec 1 (list): only GeneralValue objects or Type 3 objects.
    Spec 2 (dict): only one key-value pair, i.e. len(list(hierarchy_spec.items())) == 1.
                   for the key-value pair, key must be a GeneralValue object but value can be Type 1,2 or 3 objects.
    Spec 3 (tuple): a 2-tuple, where the left entry can be anything but right entry must be Type 1,2 or 3

    Example of a valid hierarchy specification: (used in later comments)
    {
        name: {
            subject: [
                rank,
                grade
            ]
        }
    }
    (where rank has name 'Rank', and grade has name 'Grade')

    Example of a dictionary with respect to that hierarchy:
    {
        'Alice': {
            'Math': {
                'Rank': 1,
                'Grade': 4.0
            },
            'CS': {
                'Rank': 2,
                'Grade': 3.9
            }
        },
        'Bob': {
            'Math': {
                'Rank': 2,
                'Grade': 3.9
            },
            'CS': {
                'Rank': 1,
                'Grade': 4.0
        }
    }

    Some terms used in comments:
        "Leaf": the values at the bottom, like grade, rank above.
        "Node": the values not at the bottom, like name, subject above.
        "Branch": a root-to-leaves path, like {'Name': 'Alice', 'Subject': 'Math', 'Rank': 1, 'Grade': 4.0}
    """

    @classmethod
    def check(cls, hierarchy_spec):
        """Check whether the input is a valid hierarchy specification.

        :param hierarchy_spec: an input to be checked.
        :raise some error if hierarchy is not invalid.
        """
        try:
            if type(hierarchy_spec) is list:  # Type 1
                for value in hierarchy_spec:
                    if not isinstance(value, GeneralValue):
                        # See whether is Type 3 and check next layer
                        _, next_layer = hierarchy_spec
                        cls.check(next_layer)

            elif type(hierarchy_spec) is dict:  # Type 2
                items = list(hierarchy_spec.items())
                if len(items) != 1:
                    raise ValueError
                else:
                    _, next_layer = items[0]
                    cls.check(next_layer)
            else:  # Type 3
                _, next_layer = hierarchy_spec
                cls.check(next_layer)
        except:
            print("Invalid hierarchy specification")
            raise

    @classmethod
    def apply_post_map(cls, hierarchy_spec, value_hierarchy, state=State({})):
        """Apply post_map to every value at the "leaf" of hierarchy.

        :param hierarchy_spec: a valid hierarchy (or sub-hierarchy).
        :param state (State): a state maybe used in post_map.
        :param value_hierarchy (dict): a dictionary to apply post_map with respect to hierarchy_spec.

        Example:
            Suppose post_map for both rank and grade is +1
            (value hierarchy) before:
                {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}}}

            (value hierarchy) after:
                {'Alice': {'Math': {'Rank': 2, 'Grade': 5.0}}}
        """
        if type(hierarchy_spec) is list:  # Type 1

            for value in hierarchy_spec:
                if isinstance(value, GeneralValue):
                    # Type 1: at the "leaf" of hierarchy
                    val_name = value.name
                    val = value.post_map_f(state, value_hierarchy[val_name])
                    value_hierarchy[val_name] = val
                else:
                    # Type 3: at the "node" of hierarchy, go to next hierarchy_spec
                    val, next_layer = value
                    cls.apply_post_map(next_layer, value_hierarchy[val], state)

        elif type(hierarchy_spec) is dict:  # Type 2: at the "node" of hierarchy, go to next hierarchy_spec
            value, next_layer = list(hierarchy_spec.items())[0]
            for val, _ in value_hierarchy.items():
                cls.apply_post_map(next_layer, value_hierarchy[val], state)

        else:  # Type 3: at the "node" of hierarchy, go to next hierarchy_spec
            val, next_layer = hierarchy_spec
            cls.apply_post_map(next_layer, value_hierarchy[val], state)

    @classmethod
    def traverse(cls, hierarchy_spec, line, state, value_hierarchy):
        """Store the values in the line in the input dictionary by traversing the hierarchy.

        :param hierarchy_spec: a valid hierarchy (or sub-hierarchy).
        :param line (Line or dict): the current line read.
        :param current_state (State or dict): the current state maybe used in map.
        :param value_hierarchy (dict): the dictionary to store values.

        Example:
            Read (line):
                "Alice,Math,1,4.0"
            Add to (value_hierarchy) and merge if necessary:
                {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}}}

        """
        if type(hierarchy_spec) is list:  # Type 1

            for value in hierarchy_spec:

                if isinstance(value, StateValue):
                    # get value from state and apply map
                    val_name = value.name
                    val = value.map_f(state, state[val_name])
                elif isinstance(value, Value):
                    # get value from line and apply map
                    val_name = value.name
                    val = value.map_f(state, line[val_name])
                else:
                    # Type 3: store value and go to next layer
                    val, next_layer = value
                    value_hierarchy[val] = {}
                    cls.traverse(next_layer, line, state, value_hierarchy[val])

                # if another value is already stored, apply reduce and store the value in dictionary
                if val_name in value_hierarchy:
                    current_val = value_hierarchy[val_name]
                    reduced_val = value.reduce_f(current_val, val)
                    value_hierarchy[val_name] = reduced_val
                else:
                    value_hierarchy[val_name] = val

        elif type(hierarchy_spec) is dict:  # Type 2

            value, next_layer = list(hierarchy_spec.items())[0]

            if isinstance(value, StateValue):
                # get value from state and apply map
                val_name = value.name
                val = value.map_f(state, state[val_name])
            elif isinstance(value, Value):
                # get value from line and apply map
                val_name = value.name
                val = value.map_f(state, line[val_name])
            else:
                raise ValueError("Invalid hierarchy specification.")

            if val in value_hierarchy:
                # if value already stored, go to next layer and merge with current hierarchy
                new_hierarchy = {}
                cls.traverse(next_layer, line, state, new_hierarchy)
                current_hierarchy = value_hierarchy[val]
                merged_hierarchy = cls.merge(next_layer, current_hierarchy, new_hierarchy)
                value_hierarchy[val] = merged_hierarchy

            else:
                # store value and go to next layer
                value_hierarchy[val] = {}
                cls.traverse(next_layer, line, state, value_hierarchy[val])

        else:  # Type 3: store value and go to next layer
            val, next_layer = hierarchy_spec
            value_hierarchy[val] = {}
            cls.traverse(next_layer, line, state, value_hierarchy[val])

    @classmethod
    def merge(cls, hierarchy_spec, value_hierarchy1, value_hierarchy2, post=False, state=State({}), top=True):
        """Merge two value hierarchies according to a hierarchy specification.

        :param hierarchy_spec: a valid hierarchy specification.
        :param value_hierarchy1 (dict): a dictionary with respect to hierarchy_spec.
        :param value_hierarchy2 (dict): a dictionary with respect to hierarchy_spec.
        :param post (bool): apply map, reduce or apply post_map, post_reduce
        :param state (State): the current state maybe used in map or post_map
        :param top (bool): whether to copy the input (value_hierarchy1)
        :return (dict): the merged value hierarchy

        Example 1:
            (value_hierarchy1):
                {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}}}
            (value_hierarchy2):
                {'Alice': {'CS': {'Rank': 2, 'Grade': 3.9}}}

            Merge:
                {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}, 'CS': {'Rank': 2, 'Grade': 3.9}}}

        Example 2: (Suppose reduce_f of grade is max, reduce_f of rank is min)
            (value_hierarchy1):
                {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}}}
            (value_hierarchy2):
                {'Alice': {'Math': {'Rank': 2, 'Grade': 3.9}}}

            Merge:
                {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}}

        """
        if top:  # prevent contamination of the top level input dictionary
            merged_hierarchy = deepcopy(value_hierarchy1)
        else:
            merged_hierarchy = value_hierarchy1.copy()
        if top and post:  # apply post map at the top level
            cls.apply_post_map(hierarchy_spec, merged_hierarchy, state)

        if type(hierarchy_spec) is list:  # Type 1

            for value in hierarchy_spec:

                if isinstance(value, GeneralValue):
                    # apply reduce (or post_map, post_reduce) to "leaf" values
                    val1 = merged_hierarchy[value.name]
                    val2 = value_hierarchy2[value.name]
                    if post:
                        new_val1 = val1  # already apply post map
                        new_val2 = value.post_map_f(state, val2)
                        new_val = value.post_reduce_f(new_val1, new_val2)
                    else:
                        new_val = value.reduce_f(val1, val2)
                    merged_hierarchy[value.name] = new_val

                else:  # Type 3: at "node", merge next layer
                    val, next_layer = value
                    merged_hierarchy[val] = \
                        cls.merge(next_layer, merged_hierarchy[val], value_hierarchy2[val], post, state, False)

        elif type(hierarchy_spec) is dict:  # Type 2: at "node", merge every sub-hierarchy
            _, next_layer = list(hierarchy_spec.items())[0]
            for val, next_val_layer2 in value_hierarchy2.items():
                if val in merged_hierarchy:
                    next_val_layer1 = merged_hierarchy[val]
                    merged_hierarchy[val] = \
                        cls.merge(next_layer, next_val_layer1, next_val_layer2, post, state, False)
                else:
                    if post:
                        cls.apply_post_map(next_layer, next_val_layer2, state)
                    merged_hierarchy[val] = next_val_layer2

        else:  # Type 3: at "node", merge next layer
            val, next_layer = hierarchy_spec
            merged_hierarchy[val] = \
                cls.merge(next_layer, merged_hierarchy[val], value_hierarchy2[val], post, state, False)

        return merged_hierarchy

    @classmethod
    def flatten(cls, hierarchy_spec, value_hierarchy):
        """Convert a value hierarchy to all "branches" with respect to a hierarchy specification.

        :param hierarchy_spec: a valid hierarchy specification.
        :param value_hierarchy (dict): a dictionary with respect to hierarchy_spec.

        Example:
            Convert (value_hierarchy):
                {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}, 'CS': {'Rank': 2, 'Grade': 3.9}}}

            To (value_lines):
                [
                    {'Name': 'Alice', 'Subject': 'Math', 'Rank': 1, 'Grade': 4.0},
                    {'Name': 'Alice', 'Subject': 'CS', 'Rank': 2, 'Grade': 3.9}
                ]
        """
        value_lines = []

        def flatten_helper(hierarchy_spec, value_hierarchy, value_lines, value_line={}):
            if type(hierarchy_spec) is list:  # Type 1
                current_value_line = value_line
                for value in hierarchy_spec:
                    if isinstance(value, GeneralValue):
                        # at "leaf", store (value name => value) in current value line
                        current_value_line[value.name] = value_hierarchy[value.name]
                    else:  # Type 3: at "node", go to next layer
                        val, next_layer = value
                        flatten_helper(next_layer, value_hierarchy[val], value_lines, current_value_line)
                value_lines.append(current_value_line)  # finish a "branch"

            elif type(hierarchy_spec) is dict:  # Type 2: at "node"
                for root, next in value_hierarchy.items():
                    current_value_line = value_line.copy()  # copy the current value line to different sub-hierarchies
                    value, next_layer = list(hierarchy_spec.items())[0]
                    current_value_line[value.name] = root  # store the current "node"
                    flatten_helper(next_layer, next, value_lines, current_value_line)  # flatten next layer

            else:  # Type 3: at "node", go to next layer
                current_value_line = value_line
                val, next_layer = hierarchy_spec
                flatten_helper(next_layer, value_hierarchy[val], value_lines, current_value_line)

        flatten_helper(hierarchy_spec, value_hierarchy, value_lines)
        return value_lines

    @classmethod
    def traverse_lines(cls, hierarchy_spec, value_lines, state=State({}), filter_f=lambda _, x: True):
        """Covert a list of lines to a dictionary with respect to hierarchy spec (backward of flatten).

        :param hierarchy_spec: a valid hierarchy specification
        :param value_lines ([dict]): a list of (name => value) dictionary
        :param state (State or dict): the evolving state used in map/post_map/filter
        :param filter_f (State * Line -> bool): the predicate for filtering lines with access to state.
        :return: the result dictionary
        """
        value_hierarchy = {}

        for line in value_lines:

            # update state
            state.release()
            state.update(line)
            state.lock()

            # filter out a line if predicate returns false
            if not filter_f(state, line):
                continue

            # traverse a line with respect to hierarchy
            cls.traverse(hierarchy_spec, line, state, value_hierarchy)

        return value_hierarchy



from __future__ import absolute_import
from __future__ import print_function

from uniRW.Value import GeneralValue
from uniRW.deprecated.Key import GeneralKey


class Writer:

    def __init__(self, KeyValues):
        self.KeyValues = KeyValues

    def write(self, out_file, key_val_dict, mode='w', sort_by=None, reverse=False):
        output = open(out_file.file_name, mode)
        header_line = out_file.line.get_line(out_file.header)
        for foreword_line in out_file.foreword:
            print(foreword_line, file=output)
        print(header_line, file=output, end='')

        if sort_by==None:
            sorted_items = sorted(key_val_dict.items(), reverse=reverse)
        elif sort_by=='key':
            sorted_items = sorted(key_val_dict.items(), key= lambda kv: kv[0], reverse=reverse)
        else:
            sorted_items = sorted(key_val_dict.items(), key= lambda kv: kv[1][sort_by], reverse=reverse)

        for key, val_dict in sorted_items:

            values = []

            for var in self.KeyValues:
                if isinstance(var, GeneralKey):
                    values.append(var.to_string(key))
                elif isinstance(var, GeneralValue):
                    values.append(var.to_string(val_dict[var.name]))

            print(out_file.line.get_line(values), file=output, end='')

        for epilogue_line in out_file.epilogue:
            print(epilogue_line, file=output)

        output.close()



from __future__ import absolute_import

import unittest

from uniRW.State import State


class TestState(unittest.TestCase):

    def test_get(self):
        for i in range(100):
            state = State({'count': i, 'count+1': i + 1, 'count*2': i * 2})
            self.assertEqual(state['count'], i)
            self.assertEqual(state['count+1'], i + 1)
            self.assertEqual(state['count*2'], i * 2)
            with self.assertRaises(KeyError):
                print(state['count' + str(i)])

    def test_set(self):
        state = State({'count': 0, 'count+1': 0, 'count*2': 0})
        with self.assertRaises(KeyError):
            state['count'] = 1
        state.release()
        for i in range(100):
            state['count'] = i
            state['count+1'] = i + 1
            state['count*2'] = i * 2
            self.assertEqual(state['count'], i)
            self.assertEqual(state['count+1'], i + 1)
            self.assertEqual(state['count*2'], i * 2)
        state.lock()
        with self.assertRaises(KeyError):
            state['count'] = 0

    def test_update(self):
        state = State({'count1': 0, 'count2': 0, 'count3': 0})
        count1 = 0
        count2 = 0

        def update1(state, _):
            state['count1'] += 1

        def update2(state, _):
            state['count2'] += 2

        def update3(state, val):
            state['count3'] = val

        state.update_func = update1
        with self.assertRaises(KeyError):
            state.update(0)

        state.release()
        for i in range(100):
            count1 += 1
            state.update(0)
            self.assertEqual(state['count1'], count1)
            self.assertEqual(state['count2'], 0)
            self.assertEqual(state['count3'], 0)
        state.lock()

        state.update_func = update2
        with self.assertRaises(KeyError):
            state.update(0)

        state.release()
        for i in range(100):
            count2 += 2
            state.update(0)
            self.assertEqual(state['count1'], count1)
            self.assertEqual(state['count2'], count2)
            self.assertEqual(state['count3'], 0)
        state.lock()

        state.update_func = update3
        with self.assertRaises(KeyError):
            state.update(0)

        state.release()
        for i in range(100):
            state.update(0)
            self.assertEqual(state['count1'], count1)
            self.assertEqual(state['count2'], count2)
            self.assertEqual(state['count3'], 0)
        state.lock()


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import

import unittest

from uniRW.Hierarchy import Hierarchy
from uniRW.Value import Value


class TestHierarchy(unittest.TestCase):

    def setUp(self):
        self.name = Value('Name')
        self.subject = Value('Subject')
        self.grade = Value('Grade')
        self.rank = Value('Rank')

    def test_check(self):
        name = self.name
        subject = self.subject
        grade = self.grade
        rank = self.rank
        s_name = self.name.state_value()
        s_subject = self.subject.state_value()
        s_grade = self.grade.state_value()
        s_rank = self.rank.state_value()

        Hierarchy.check({name: [grade]})
        Hierarchy.check({name: [grade, subject]})
        Hierarchy.check({name: [grade, subject, rank]})
        Hierarchy.check({name: {subject: [grade, rank]}})
        Hierarchy.check({name: {subject: {grade: [rank]}}})
        Hierarchy.check({s_name: [s_grade]})
        Hierarchy.check({s_name: [s_grade, s_subject]})
        Hierarchy.check({s_name: [s_grade, s_subject, s_rank]})
        Hierarchy.check({s_name: {s_subject: [s_grade, s_rank]}})
        Hierarchy.check({s_name: {s_subject: {s_grade: [s_rank]}}})

        with self.assertRaises(Exception):
            Hierarchy.check(name)
        with self.assertRaises(Exception):
            Hierarchy.check({name: grade})
        with self.assertRaises(Exception):
            Hierarchy.check({name: {subject: [grade], rank: [grade]}})
        with self.assertRaises(Exception):
            Hierarchy.check([{name: grade}])
        with self.assertRaises(Exception):
            Hierarchy.check(s_name)
        with self.assertRaises(Exception):
            Hierarchy.check({s_name: s_grade})
        with self.assertRaises(Exception):
            Hierarchy.check({s_name: {s_subject: [s_grade], s_rank: [s_grade]}})
        with self.assertRaises(Exception):
            Hierarchy.check([{s_name: s_grade}])

    def test_apply_post_map(self):
        name = self.name
        subject = self.subject
        grade = self.grade
        rank = self.rank

        grade.post_map_f = lambda _, x: x - 1.0
        rank.post_map_f = lambda _, x: x + 1

        hierarchy1 = {name: [grade]}
        hierarchy2 = {name: {subject: [rank, grade]}}

        value_dict1 = {'Alice': {'Grade': 4.0}, 'Bob': {'Grade': 3.0}}
        value_dict2 = {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}},
                       'Bob': {'CS': {'Rank': 2, 'Grade': 3.0}}}

        Hierarchy.apply_post_map(hierarchy1, value_dict1, {})
        Hierarchy.apply_post_map(hierarchy2, value_dict2, {})

        self.assertEqual(value_dict1, {'Alice': {'Grade': 3.0}, 'Bob': {'Grade': 2.0}})
        self.assertEqual(value_dict2, {'Alice': {'Math': {'Rank': 2, 'Grade': 3.0}},
                                       'Bob': {'CS': {'Rank': 3, 'Grade': 2.0}}})

    def test_merge(self):
        name = self.name
        subject = self.subject
        grade = self.grade
        rank = self.rank.state_value()

        grade.reduce_f = max
        rank.reduce_f = min
        grade.post_map_f = lambda _, x: x - 1.0
        rank.post_map_f = lambda _, x: x + 1
        grade.post_reduce_f = min
        rank.post_reduce_f = max

        hierarchy1 = {name: [grade]}
        hierarchy2 = {name: {subject: [rank, grade]}}

        value_dict11 = {'Alice': {'Grade': 4.0}}
        value_dict12 = {'Alice': {'Grade': 3.0}}
        value_dict13 = {'Bob': {'Grade': 3.5}}
        value_dict20 = {'Alice': {'Math': {'Rank': 2, 'Grade': 3.0}}}
        value_dict21 = {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}}}
        value_dict22 = {'Alice': {'CS': {'Rank': 2, 'Grade': 3.9}}}
        value_dict23 = {'Bob': {'Math': {'Rank': 2, 'Grade': 3.0}}}

        self.assertEqual(Hierarchy.merge(hierarchy1, value_dict11, value_dict12),
                         {'Alice': {'Grade': 4.0}})
        self.assertEqual(Hierarchy.merge(hierarchy1, value_dict11, value_dict12, True),
                         {'Alice': {'Grade': 2.0}})
        self.assertEqual(Hierarchy.merge(hierarchy1, value_dict11, value_dict13),
                         {'Alice': {'Grade': 4.0}, 'Bob': {'Grade': 3.5}})
        self.assertEqual(Hierarchy.merge(hierarchy1, value_dict11, value_dict13, True),
                         {'Alice': {'Grade': 3.0}, 'Bob': {'Grade': 2.5}})

        self.assertEqual(Hierarchy.merge(hierarchy2, value_dict20, value_dict21),
                         {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}}})
        self.assertEqual(Hierarchy.merge(hierarchy2, value_dict20, value_dict21, True),
                         {'Alice': {'Math': {'Rank': 3, 'Grade': 2.0}}})
        self.assertEqual(Hierarchy.merge(hierarchy2, value_dict21, value_dict22),
                         {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0},
                                    'CS': {'Rank': 2, 'Grade': 3.9}}})
        self.assertEqual(Hierarchy.merge(hierarchy2, value_dict21, value_dict22, True),
                         {'Alice': {'Math': {'Rank': 2, 'Grade': 3.0},
                                    'CS': {'Rank': 3, 'Grade': 2.9}}})
        self.assertEqual(Hierarchy.merge(hierarchy2, value_dict21, value_dict23),
                         {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}},
                          'Bob': {'Math': {'Rank': 2, 'Grade': 3.0}}})
        self.assertEqual(Hierarchy.merge(hierarchy2, value_dict21, value_dict23, True),
                         {'Alice': {'Math': {'Rank': 2, 'Grade': 3.0}},
                          'Bob': {'Math': {'Rank': 3, 'Grade': 2.0}}})

    def test_traverse(self):
        name = self.name
        subject = self.subject
        grade = self.grade
        rank = self.rank.state_value()

        grade.map_f = lambda _, x: float(x)
        grade.reduce_f = max
        rank.map_f = lambda _, x: int(x)
        rank.reduce_f = min

        hierarchy1 = {name: [grade]}
        hierarchy2 = {name: {subject: [rank, grade]}}

        line11 = {'Name': 'Alice', 'Subject': 'Math',  'Grade': '3.0'}
        line12 = {'Name': 'Alice', 'Subject': 'Math',  'Grade': '4.0'}
        line13 = {'Name': 'Alice', 'Subject': 'CS', 'Grade': '3.5'}
        line21 = {'Name': 'Bob', 'Subject': 'CS',  'Grade': '3.5'}
        line22 = {'Name': 'Bob', 'Subject': 'CS', 'Grade': '3.0'}
        line23 = {'Name': 'Bob', 'Subject': 'Math', 'Grade': '4.0'}

        value_dict1 = {}
        Hierarchy.traverse(hierarchy1, line11, {'Rank': '3'}, value_dict1)
        self.assertEqual(value_dict1, {'Alice': {'Grade': 3.0}})
        Hierarchy.traverse(hierarchy1, line12, {'Rank': '1'}, value_dict1)
        self.assertEqual(value_dict1, {'Alice': {'Grade': 4.0}})
        Hierarchy.traverse(hierarchy1, line13, {'Rank': '2'}, value_dict1)
        self.assertEqual(value_dict1, {'Alice': {'Grade': 4.0}})
        Hierarchy.traverse(hierarchy1, line21, {'Rank': '2'}, value_dict1)
        self.assertEqual(value_dict1, {'Alice': {'Grade': 4.0}, 'Bob': {'Grade': 3.5}})
        Hierarchy.traverse(hierarchy1, line22, {'Rank': '3'}, value_dict1)
        self.assertEqual(value_dict1, {'Alice': {'Grade': 4.0}, 'Bob': {'Grade': 3.5}})
        Hierarchy.traverse(hierarchy1, line23, {'Rank': '1'}, value_dict1)
        self.assertEqual(value_dict1, {'Alice': {'Grade': 4.0}, 'Bob': {'Grade': 4.0}})

        value_dict2 = {}
        Hierarchy.traverse(hierarchy2, line11, {'Rank': '3'}, value_dict2)
        self.assertEqual(value_dict2, {'Alice': {'Math': {'Grade': 3.0, 'Rank': 3}}})
        Hierarchy.traverse(hierarchy2, line12, {'Rank': '1'}, value_dict2)
        self.assertEqual(value_dict2, {'Alice': {'Math': {'Grade': 4.0, 'Rank': 1}}})
        Hierarchy.traverse(hierarchy2, line13, {'Rank': '2'}, value_dict2)
        self.assertEqual(value_dict2, {'Alice': {'Math': {'Grade': 4.0, 'Rank': 1},
                                                 'CS': {'Grade': 3.5, 'Rank': 2}}})
        Hierarchy.traverse(hierarchy2, line21, {'Rank': '2'}, value_dict2)
        self.assertEqual(value_dict2, {'Alice': {'Math': {'Grade': 4.0, 'Rank': 1},
                                                 'CS': {'Grade': 3.5, 'Rank': 2}},
                                       'Bob': {'CS': {'Grade': 3.5, 'Rank': 2}}})
        Hierarchy.traverse(hierarchy2, line22, {'Rank': '3'}, value_dict2)
        self.assertEqual(value_dict2, {'Alice': {'Math': {'Grade': 4.0, 'Rank': 1},
                                                 'CS': {'Grade': 3.5, 'Rank': 2}},
                                       'Bob': {'CS': {'Grade': 3.5, 'Rank': 2}}})
        Hierarchy.traverse(hierarchy2, line23, {'Rank': '1'}, value_dict2)
        self.assertEqual(value_dict2, {'Alice': {'Math': {'Grade': 4.0, 'Rank': 1},
                                                 'CS': {'Grade': 3.5, 'Rank': 2}},
                                       'Bob': {'CS': {'Grade': 3.5, 'Rank': 2},
                                               'Math': {'Grade': 4.0, 'Rank': 1}}})

    def test_flatten(self):
        name = self.name
        subject = self.subject
        grade = self.grade
        rank = self.rank.state_value()

        hierarchy1 = {name: [grade]}
        hierarchy2 = {name: {subject: [rank, grade]}}

        value_dict11 = {'Alice': {'Grade': 4.0}}
        value_dict12 = {'Alice': {'Grade': 4.0}, 'Bob': {'Grade': 3.0}}
        value_dict21 = {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}, 'CS': {'Rank': 2, 'Grade': 3.9}}}
        value_dict22 = {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}, 'CS': {'Rank': 2, 'Grade': 3.8}},
                        'Bob': {'Math': {'Rank': 2, 'Grade': 3.0}, 'CS': {'Rank': 1, 'Grade': 3.9}}}

        def key(v):
            return v['Grade']

        value_lines11 = Hierarchy.flatten(hierarchy1, value_dict11)
        self.assertEqual(value_lines11, [{'Name': 'Alice', 'Grade': 4.0}])

        value_lines12 = Hierarchy.flatten(hierarchy1, value_dict12)
        self.assertEqual(sorted(value_lines12, key=key),
                         sorted([{'Name': 'Alice', 'Grade': 4.0}, {'Name': 'Bob', 'Grade': 3.0}], key=key))

        value_lines21 = Hierarchy.flatten(hierarchy2, value_dict21)
        self.assertEqual(sorted(value_lines21, key=key),
                         sorted([{'Name': 'Alice', 'Subject': 'Math', 'Rank': 1, 'Grade': 4.0},
                                 {'Name': 'Alice', 'Subject': 'CS', 'Rank': 2, 'Grade': 3.9}], key=key))

        value_lines22 = Hierarchy.flatten(hierarchy2, value_dict22)
        self.assertEqual(sorted(value_lines22, key=key),
                         sorted([{'Name': 'Alice', 'Subject': 'Math', 'Rank': 1, 'Grade': 4.0},
                                 {'Name': 'Alice', 'Subject': 'CS', 'Rank': 2, 'Grade': 3.8},
                                 {'Name': 'Bob', 'Subject': 'Math', 'Rank': 2, 'Grade': 3.0},
                                 {'Name': 'Bob', 'Subject': 'CS', 'Rank': 1, 'Grade': 3.9}], key=key))

    def test_traverse_lines(self):
        name = self.name
        subject = self.subject
        grade = self.grade
        rank = self.rank

        hierarchy1 = {name: [grade]}
        hierarchy2 = {name: {subject: [rank, grade]}}

        value_dict11 = {'Alice': {'Grade': 4.0}}
        value_dict12 = {'Alice': {'Grade': 4.0}, 'Bob': {'Grade': 3.0}}
        value_dict21 = {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}, 'CS': {'Rank': 2, 'Grade': 3.9}}}
        value_dict22 = {'Alice': {'Math': {'Rank': 1, 'Grade': 4.0}, 'CS': {'Rank': 2, 'Grade': 3.8}},
                        'Bob': {'Math': {'Rank': 2, 'Grade': 3.0}, 'CS': {'Rank': 1, 'Grade': 3.9}}}

        value_lines11 = [{'Name': 'Alice', 'Grade': 4.0}]
        value_lines12 = [{'Name': 'Alice', 'Grade': 4.0}, {'Name': 'Bob', 'Grade': 3.0}]
        value_lines21 = [{'Name': 'Alice', 'Subject': 'Math', 'Rank': 1, 'Grade': 4.0},
                         {'Name': 'Alice', 'Subject': 'CS', 'Rank': 2, 'Grade': 3.9}]
        value_lines22 = [{'Name': 'Alice', 'Subject': 'Math', 'Rank': 1, 'Grade': 4.0},
                         {'Name': 'Alice', 'Subject': 'CS', 'Rank': 2, 'Grade': 3.8},
                         {'Name': 'Bob', 'Subject': 'Math', 'Rank': 2, 'Grade': 3.0},
                         {'Name': 'Bob', 'Subject': 'CS', 'Rank': 1, 'Grade': 3.9}]

        self.assertEqual(value_dict11, Hierarchy.traverse_lines(hierarchy1, value_lines11))
        self.assertEqual(value_dict12, Hierarchy.traverse_lines(hierarchy1, value_lines12))
        self.assertEqual(value_dict21, Hierarchy.traverse_lines(hierarchy2, value_lines21))
        self.assertEqual(value_dict22, Hierarchy.traverse_lines(hierarchy2, value_lines22))

if __name__ == '__main__':
    unittest.main()





def check_and_apply_f(f_dict, f, key, arg):
    return f_dict[key](arg) if key in f_dict else f(arg)


def check_and_apply_2(f_dict, key, arg1, arg2):
    return f_dict[key](arg1, arg2) if key in f_dict else arg2



from __future__ import absolute_import, print_function

from uniRW.Value import GeneralValue
from uniRW.Hierarchy import Hierarchy


class HWriter:

    def __init__(self, hierarchy_spec, value_line):
        """Initialize the writer.

        :param hierarchy_spec: a hierarchy specification (described in Hierarchy.py) describing the input dictionary.
        :param value_line ([GeneralValue]): an ordered value object list specifying each line of output.
        """
        Hierarchy.check(hierarchy_spec)
        for value in value_line:
            if not isinstance(value, GeneralValue):
                raise ValueError("Value is not a value object")

        self.hierarchy_spec = hierarchy_spec
        self.value_line = value_line

    def write(self, out_file, value_hierarchy, mode='w', sort_by=None, reverse=False):
        """Write a file according to value line specification.

        :param out_file (OutputFile): the file to write.
        :param value_hierarchy (dict): a value hierarchy (dictionary) for writing to the file.
        :param mode (str): writing mode ("w", "w+", ...)
        :param sort_by (str): the name of a value to sort by.
        :param reverse: sort in reverse order or not.
        """

        # flatten the value hierarchy (dictionary)
        value_lines = Hierarchy.flatten(self.hierarchy_spec, value_hierarchy)
        self.write_lines(out_file, value_lines, mode, sort_by, reverse)

    def write_lines(self, out_file, value_lines, mode, sort_by, reverse):
        """Write a file according to value specification, with a flatten value hierarchy.
        """
        output = open(out_file.file_name, mode)

        # print foreword
        for foreword_line in out_file.foreword:
            print(foreword_line, file=output)

        # print header
        if out_file.header:
            header_line = out_file.line.get_line(out_file.header)
        else:
            header_line = out_file.line.get_line([value.name for value in self.value_line])
        print(header_line, file=output, end='')

        # sort the lines
        if not sort_by:
            sorted_lines = value_lines
        else:
            sorted_lines = sorted(value_lines, key=lambda v_dict: v_dict[sort_by], reverse=reverse)

        # print value with respect to value line specification
        for line in sorted_lines:
            values = []
            for value in self.value_line:
                values.append(value.to_string(line[value.name]))
            print(out_file.line.get_line(values), file=output, end='')

        # print the epilogue
        for epilogue_line in out_file.epilogue:
            print(epilogue_line, file=output)

        output.close()



from __future__ import absolute_import

from uniRW.util import *
from uniRW.File import DataFile, OutputFile, File
from uniRW.Line import Line, OutputLine
from uniRW.State import State
from uniRW.Value import Value, StateValue, GeneralValue
from uniRW.Hierarchy import Hierarchy
from uniRW.HReader import HReader
from uniRW.HWriter import HWriter

# backward compatibility
from uniRW.deprecated.Key import Key, StateKey, GeneralKey
from uniRW.deprecated.RW import read, readAll, write
from uniRW.deprecated.Reader import Reader
from uniRW.deprecated.Writer import Writer



from __future__ import absolute_import

import unittest

from uniRW.Line import Line, OutputLine


class TestLine(unittest.TestCase):

    def setUp(self):
        self.comma_line = Line(',')
        self.space_line = Line('\s+')
        self.data_comma_line1 = 'A,B,C,D'
        self.data_comma_line2 = '1,2,3,4'
        self.data_space_line1 = 'A  B C   D'
        self.data_space_line2 = '1  2 3   4'

    def test_header(self):
        self.comma_line.store(self.data_comma_line1)
        self.comma_line.set_header()
        self.assertEqual(self.comma_line.name_col_dict, {'A': 0, 'B': 1, 'C': 2, 'D': 3})
        self.comma_line.clear()
        self.assertEqual(self.comma_line.name_col_dict, {})

        self.comma_line.store(self.data_comma_line2)
        self.comma_line.set_header()
        self.assertEqual(self.comma_line.name_col_dict, {'1': 0, '2': 1, '3': 2, '4': 3})
        self.comma_line.clear()
        self.assertEqual(self.comma_line.name_col_dict, {})

        self.space_line.store(self.data_space_line1)
        self.space_line.set_header()
        self.assertEqual(self.space_line.name_col_dict, {'A': 0, 'B': 1, 'C': 2, 'D': 3})
        self.space_line.clear()
        self.assertEqual(self.space_line.name_col_dict, {})

        self.space_line.store(self.data_space_line2)
        self.space_line.set_header()
        self.assertEqual(self.space_line.name_col_dict, {'1': 0, '2': 1, '3': 2, '4': 3})
        self.space_line.clear()
        self.assertEqual(self.space_line.name_col_dict, {})

    def test_get(self):
        self.comma_line.store(self.data_comma_line1)
        self.comma_line.set_header()
        self.comma_line.store(self.data_comma_line2)

        self.assertEqual(self.comma_line['A'], '1')
        self.assertEqual(self.comma_line['B'], '2')
        self.assertEqual(self.comma_line['C'], '3')
        self.assertEqual(self.comma_line['D'], '4')
        with self.assertRaises(KeyError):
            print(self.comma_line['E'])

        self.assertEqual(self.comma_line[0], '1')
        self.assertEqual(self.comma_line[1], '2')
        self.assertEqual(self.comma_line[2], '3')
        self.assertEqual(self.comma_line[3], '4')
        self.assertEqual(self.comma_line[-4], '1')
        self.assertEqual(self.comma_line[-3], '2')
        self.assertEqual(self.comma_line[-2], '3')
        self.assertEqual(self.comma_line[-1], '4')
        with self.assertRaises(IndexError):
            print(self.comma_line[4])

        self.comma_line.clear()

        self.comma_line.store(self.data_comma_line2)
        self.comma_line.set_header()
        self.comma_line.store(self.data_comma_line1)

        self.assertEqual(self.comma_line['1'], 'A')
        self.assertEqual(self.comma_line['2'], 'B')
        self.assertEqual(self.comma_line['3'], 'C')
        self.assertEqual(self.comma_line['4'], 'D')
        with self.assertRaises(KeyError):
            print(self.comma_line['5'])

        self.assertEqual(self.comma_line[0], 'A')
        self.assertEqual(self.comma_line[1], 'B')
        self.assertEqual(self.comma_line[2], 'C')
        self.assertEqual(self.comma_line[3], 'D')
        self.assertEqual(self.comma_line[-4], 'A')
        self.assertEqual(self.comma_line[-3], 'B')
        self.assertEqual(self.comma_line[-2], 'C')
        self.assertEqual(self.comma_line[-1], 'D')
        with self.assertRaises(IndexError):
            print(self.comma_line[4])
        with self.assertRaises(IndexError):
            print(self.comma_line[-5])

        self.comma_line.clear()

        self.space_line.store(self.data_space_line1)
        self.space_line.set_header()
        self.space_line.store(self.data_space_line2)

        self.assertEqual(self.space_line['A'], '1')
        self.assertEqual(self.space_line['B'], '2')
        self.assertEqual(self.space_line['C'], '3')
        self.assertEqual(self.space_line['D'], '4')
        with self.assertRaises(KeyError):
            print(self.space_line['E'])

        self.assertEqual(self.space_line[0], '1')
        self.assertEqual(self.space_line[1], '2')
        self.assertEqual(self.space_line[2], '3')
        self.assertEqual(self.space_line[3], '4')
        self.assertEqual(self.space_line[-4], '1')
        self.assertEqual(self.space_line[-3], '2')
        self.assertEqual(self.space_line[-2], '3')
        self.assertEqual(self.space_line[-1], '4')
        with self.assertRaises(IndexError):
            print(self.space_line[4])
        with self.assertRaises(IndexError):
            print(self.space_line[-5])

        self.space_line.clear()

        self.space_line.store(self.data_space_line2)
        self.space_line.set_header()
        self.space_line.store(self.data_space_line1)

        self.assertEqual(self.space_line['1'], 'A')
        self.assertEqual(self.space_line['2'], 'B')
        self.assertEqual(self.space_line['3'], 'C')
        self.assertEqual(self.space_line['4'], 'D')
        with self.assertRaises(KeyError):
            print(self.space_line['5'])

        self.assertEqual(self.space_line[0], 'A')
        self.assertEqual(self.space_line[1], 'B')
        self.assertEqual(self.space_line[2], 'C')
        self.assertEqual(self.space_line[3], 'D')
        self.assertEqual(self.space_line[-4], 'A')
        self.assertEqual(self.space_line[-3], 'B')
        self.assertEqual(self.space_line[-2], 'C')
        self.assertEqual(self.space_line[-1], 'D')
        with self.assertRaises(IndexError):
            print(self.space_line[4])
        with self.assertRaises(IndexError):
            print(self.space_line[-5])

        self.space_line.clear()

    def test_output_line(self):
        comma_line = OutputLine(',', 'a,', ',d')
        space_line = OutputLine(' ', 'd ', ' a')

        self.assertEqual(comma_line.get_line(['b', 'c']), 'a,b,c,d')
        self.assertEqual(comma_line.get_line(['1', '2', '3']), 'a,1,2,3,d')
        self.assertEqual(space_line.get_line(['b', 'c']), 'd b c a')
        self.assertEqual(space_line.get_line(['1', '2', '3']), 'd 1 2 3 a')

if __name__ == '__main__':
    unittest.main()



from distutils.core import setup

setup(
  name='uniRW',
  packages=['uniRW', 'uniRW/deprecated'],
  version='0.5.2',
  description='A universal reader and writer for stateful data file processing',
  author='Langxuan Su',
  author_email='lawliet@orion.codes',
  url='https://github.com/law-liet/uniRW',
  download_url='https://github.com/law-liet/uniRW/tarball/v0.5.2',
  keywords=['declarative', 'functional', 'reusable', 'universal', 'stateful', 'data',
              'file', 'processing', 'csv', 'reader', 'writer',  'map', 'reduce', 'filter'],
  classifiers=[
    "Programming Language :: Python",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.2",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Topic :: Text Processing",
    "Topic :: Utilities",
    "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from __future__ import absolute_import

from copy import deepcopy

from uniRW.File import DataFile
from uniRW.Hierarchy import Hierarchy
from uniRW.State import State


class HReader:

    def __init__(self, hierarchy_spec, state=State({}), filter_f=lambda _, x: True):
        """Initialize the reader.

        :param hierarchy_spec: a (or a list of) hierarchy specification (described in Hierarchy.py).
        :param state (State): the state involved in reading.
        :param filter_f (State * Line -> bool): the predicate for filtering lines with access to state.
        """
        Hierarchy.check(hierarchy_spec)
        if not isinstance(state, State):
            raise ValueError("State is a State object")

        self.hierarchy_spec = hierarchy_spec
        self.state = state
        self.filter_f = filter_f
        self.__init_state = deepcopy(self.state)

    def read_lines(self, value_lines, apply_post_map=False, carry_state=False):
        """Store the data in a list of lines with respect to the hierarchy specification of values

        :param value_lines ([dict]): a list of (name => value) dictionary
        :param apply_post_map (bool): whether apply post_map or not
        :param carry_state: whether mutate the input state
        :return (dict): the result dictionary and the final state.
        """
        # if not carry state, copy the state object such that the input state is not mutated.
        if not carry_state:
            current_state = deepcopy(self.state)
        else:
            current_state = self.state

        value_hierarchy = Hierarchy.traverse_lines(self.hierarchy_spec, value_lines, current_state, self.filter_f)

        # apply post_map_f in each value if apply_post_map is true
        if apply_post_map:
            Hierarchy.apply_post_map(self.hierarchy_spec, value_hierarchy, current_state)

        return value_hierarchy, current_state


    def read(self, data_file, mode='r', apply_post_map=False, carry_state=False):
        """Read a file to store the data with respect to the hierarchy specification of values.

        :param data_file (DataFile): the file to read.
        :param mode (str): reading mode ('r', 'r+', ...).
        :param apply_post_map (bool): whether apply post_map for each value or not.
        :param carry_state (bool): whether keep the mutated state or not.
        :return (dict, State): the result dictionary and the final state.
        """
        if not isinstance(data_file, DataFile):
            raise ValueError("Data file is not a DataFile object.")

        lineno = 0
        value_hierarchy = {}

        # if not carry state, copy the state object such that the input state is not mutated.
        if not carry_state:
            current_state = deepcopy(self.state)
        else:
            current_state = self.state

        with open(data_file.file_name, mode) as file:

            for line in file:
                try:
                    # read a line
                    data_file.line.store(line[:-1])

                    # set header if header line number matches
                    if data_file.header_lineno == lineno:
                        data_file.line.set_header()
                        lineno += 1
                        continue

                    # update state
                    current_state.release()
                    current_state.update(data_file.line)
                    current_state.lock()

                    # filter out a line if predicate returns false
                    if not self.filter_f(current_state, data_file.line):
                        lineno += 1
                        continue

                    # traverse a line with respect to hierarchy
                    Hierarchy.traverse(self.hierarchy_spec, data_file.line, current_state, value_hierarchy)

                except (KeyError, ValueError):
                    print("Error occurred when reading line" + str(lineno+1) + " of " + data_file.file_name)
                    raise

                lineno += 1

        # apply post_map_f in each value if apply_post_map is true
        if apply_post_map:
            Hierarchy.apply_post_map(self.hierarchy_spec, value_hierarchy, current_state)

        return value_hierarchy, current_state

    def read_all(self, data_files, mode='r', carry_state=False):
        """Read multiple files and store the data with respect to the hierarchy specification of values.

        :param data_files: list of files
        :return: the result dictionary
        """
        final_value_hierarchy = {}

        # read each file and merge the result dictionary
        for data_file in data_files:
            value_hierarchy, final_state = self.read(data_file, mode, carry_state)
            final_value_hierarchy = \
                Hierarchy.merge(self.hierarchy_spec, final_value_hierarchy, value_hierarchy, True, final_state, False)

        return final_value_hierarchy

    def clear_state(self):
        self.state = deepcopy(self.__init_state)



from __future__ import absolute_import

from uniRW.util import idenR


class GeneralValue:

    def __init__(self, name, to_string, map_f, reduce_f, post_map_f, post_reduce_f):
        """Initialize a value structure defining what and how to read and write

        :param name (str): the name (column name in header, or name of a state variable) of a value.
        :param to_string ('a -> str): a function turning a value to string for printing.
        :param map_f (State * 'a -> 'b): a function transforming a value (with access to state) during reading.
        :param reduce_f ('b * 'b -> 'b): an associative function reducing two values into one during reading.
        :param post_map_f (State * 'a -> 'b): a function transforming a value (with access to state) after reading.
        :param post_reduce_f ('b * 'b -> 'b): an associative function reducing two values into one after reading.
        """
        self.name = name
        self.map_f = map_f
        self.reduce_f = reduce_f
        self.to_string = to_string
        self.post_map_f = post_map_f
        self.post_reduce_f = post_reduce_f


class Value(GeneralValue):

    def __init__(self, name, to_string=str, map_f=idenR,
                 reduce_f=idenR, post_map_f=idenR, post_reduce_f=idenR):
        """Initialize a value from a line
        Subclass of GeneralValue with default input.
        """
        GeneralValue.__init__(self, name, to_string, map_f, reduce_f, post_map_f, post_reduce_f)

    def state_value(self):
        return StateValue(self.name, self.to_string, self.map_f, self.reduce_f, self.post_map_f, self.post_reduce_f)


class StateValue(GeneralValue):

    def __init__(self, name, to_string=str, map_f=idenR,
                 reduce_f=idenR, post_map_f=idenR, post_reduce_f=idenR):
        """Initialize a value from state
        Subclass of GeneralValue with default input.
        """
        GeneralValue.__init__(self, name, to_string,  map_f, reduce_f, post_map_f, post_reduce_f)

    def value(self):
        return Value(self.name, self.to_string, self.map_f, self.reduce_f, self.post_map_f, self.post_reduce_f)



# A Universal Reader and Writer
# Author: Langxuan Su
from __future__ import absolute_import
from __future__ import print_function
from re import split as resplit
from uniRW.deprecated.util import check_and_apply_f, check_and_apply_2
import sys

def read(file_name, mode, key_col, val_cols, split_re, has_header=False,
         header_dict={}, ignore_chars=[], map_fs={}, reduce_fs={},
         filter_f=None, state={}, update_state=None):

  lineno = 0
  headers = header_dict
  key_val_dict = {}
  current_state = state.copy()

  with open(file_name, mode) as file:

    for line in file:
      if line[0] in ignore_chars: continue
      a = resplit(split_re, line[:-1])

      if has_header:
        if lineno == 0:
          if headers == {}:
            headers[key_col] = a[key_col]
            for val_col in val_cols:
              headers[val_col] = a[val_col]
          lineno += 1
          continue

      if update_state != None: update_state(current_state, a)

      if filter_f != None and not filter_f(current_state, a):
        lineno += 1
        continue

      if key_col == None:
        key = check_and_apply_2(map_fs, 'key', current_state, lineno)
      else:
        key = check_and_apply_2(map_fs, 'key', current_state, a[key_col])

      for val_col in val_cols:

        val_name = headers[val_col] if val_col in headers else val_col
        if type(val_col) is str:
          if val_col in current_state:
            val = check_and_apply_2(map_fs, val_col, current_state, current_state[val_col])
          else:
            print('Error:' + val_col + ' not ' + 'in states', file=sys.stderr)
            sys.exit()
        else:
          val = check_and_apply_2(map_fs, val_col, current_state, a[val_col])

        if key in key_val_dict:
          if val_name in key_val_dict[key]:
            current_val = key_val_dict[key][val_name]
            reduced_val = check_and_apply_2(reduce_fs, val_col, current_val, val)
            key_val_dict[key][val_name] = reduced_val

          else:
            key_val_dict[key][val_name] = val
        else:
          key_val_dict[key] = {}
          key_val_dict[key][val_name] = val

      lineno += 1

  return (key_val_dict, current_state)


def readAll(file_names, mode, key_col, val_cols, split_re, has_header=False,
            header_dict={}, ignore_chars=[], map_fs={}, reduce_fs={}, filter_f=None,
            state={}, update_state=None, post_map_fs={}, post_reduce_fs={}):

  final_key_val_dict = {}

  for file_name in file_names:

    key_val_dict, final_state = read(
      file_name= file_name,
      mode= mode,
      key_col= key_col,
      val_cols= val_cols,
      split_re= split_re,
      has_header= has_header,
      header_dict= header_dict,
      ignore_chars= ignore_chars,
      map_fs= map_fs,
      reduce_fs= reduce_fs,
      filter_f= filter_f,
      state= state,
      update_state= update_state
    )

    for key, val_dict in key_val_dict.items():

      if key in final_key_val_dict:

        for val_name, raw_val in val_dict.items():

          val = check_and_apply_2(post_map_fs, val_name, final_state, raw_val)

          if val_name in final_key_val_dict[key]:

            current_val = final_key_val_dict[key][val_name]

            reduced_val = check_and_apply_2(post_reduce_fs, val_name, current_val, val)
            final_key_val_dict[key][val_name] = reduced_val
          else:
            final_key_val_dict[key][val_name] = val

      else:
        final_key_val_dict[key] = {}

        for val_name, raw_val in val_dict.items():
          val = check_and_apply_2(post_map_fs, val_name, final_state, raw_val)

          final_key_val_dict[key][val_name] = val

  return final_key_val_dict

def write(file_name, mode, key_val_dict, col_names, header, split_char,
          to_string={}, sort_by=None, foreword=[], end_char='\n', epilogue=[]):

  output = open(file_name, mode)
  header_line = split_char.join(header)
  for foreword_line in foreword:
    print(foreword_line, file=output)
  print(header_line, file=output)

  if sort_by==None:
    sorted_items = sorted(key_val_dict.items())
  elif sort_by=='key':
    sorted_items = sorted(key_val_dict.items(), key= lambda kv: kv[0])
  else:
    sorted_items = sorted(key_val_dict.items(), key= lambda kv: kv[1][sort_by])

  for key, val_dict in sorted_items:

    line = ''

    for col_name in col_names:
      if col_name=='key':
        line += check_and_apply_f(to_string, str, col_name, key)
      else:
        line += check_and_apply_f(to_string, str, col_name, val_dict[col_name])
      line += split_char

    print(line[:-1], file=output, end=end_char)

  for epilogue_line in epilogue:
    print(epilogue_line, file=output)
  output.close()
